import sys
from pathlib import Path

import pytest

import toga
import toga_dummy
from toga.icons import _APP_ICON
from toga_dummy.icons import Icon as DummyIcon

APP_RESOURCES = Path(__file__).parent / "resources"
TOGA_RESOURCES = Path(toga_dummy.__file__).parent / "resources"


class MyApp(toga.App):
    pass


@pytest.fixture
def app():
    return MyApp("Icons Test", "org.beeware.toga.icons")


@pytest.mark.parametrize(
    "path, system, sizes, extensions, final_paths",
    [
        # Relative path
        (Path("resources/red"), False, None, [".png"], APP_RESOURCES / "red.png"),
        # Absolute path (points at a file in the system resource folder,
        # but that's just because it's a location we know exists.)
        (
            TOGA_RESOURCES / "toga",
            False,
            None,
            [".png"],
            TOGA_RESOURCES / "toga.png",
        ),
        # PNG format
        ("resources/red", False, None, [".png"], APP_RESOURCES / "red.png"),
        # Explicitly specified as BMP, but ignored because platform wants PNG format
        ("resources/red.bmp", False, None, [".png"], APP_RESOURCES / "red.png"),
        # PNG format in multiple sizes
        (
            "resources/red",
            False,
            [32, 72],
            [".png"],
            {
                32: APP_RESOURCES / "red-32.png",
                72: APP_RESOURCES / "red-72.png",
            },
        ),
        # PNG format in multiple sizes, but no individually sized PNGs available
        (
            "resources/blue",
            False,
            [32, 72],
            [".png"],
            {
                32: APP_RESOURCES / "blue.png",
                72: APP_RESOURCES / "blue.png",
            },
        ),
        # Multiple formats, first option doesn't exist
        ("resources/blue", False, None, [".bmp", ".png"], APP_RESOURCES / "blue.png"),
        # Multiple formats, first match returned
        (
            "resources/orange",
            False,
            None,
            [".bmp", ".png"],
            APP_RESOURCES / "orange.bmp",
        ),
        # Relative path, platform-specific resource
        (
            Path("resources/sample"),
            False,
            None,
            [".png"],
            APP_RESOURCES / "sample-dummy.png",
        ),
        # Relative path as string, platform-specific resource
        (
            "resources/sample",
            False,
            None,
            [".png"],
            APP_RESOURCES / "sample-dummy.png",
        ),
        # Relative path, system resource
        (Path("toga"), True, None, [".png"], TOGA_RESOURCES / "toga.png"),
        # Relative path as string, system resource
        ("toga", True, None, [".png"], TOGA_RESOURCES / "toga.png"),
    ],
)
def test_create(monkeypatch, app, path, system, sizes, extensions, final_paths):
    """Icons can be created."""
    # Patch the dummy Icon class to evaluated different lookup strategies.
    monkeypatch.setattr(DummyIcon, "SIZES", sizes)
    monkeypatch.setattr(DummyIcon, "EXTENSIONS", extensions)

    # monkeypatch the current platform to report as dummy
    monkeypatch.setattr(toga.platform, "current_platform", "dummy")

    icon = toga.Icon(path, system=system)

    # Icon is bound
    assert icon._impl is not None
    # impl/interface round trips
    assert icon._impl.interface == icon

    # The icon's path is fully qualified
    assert icon._impl.path == final_paths


def test_create_fallback_missing(monkeypatch, app, capsys):
    """If a resource doesn't exist, a fallback icon is used."""
    # Prime the dummy so the app icon cannot be loaded
    monkeypatch.setattr(
        DummyIcon,
        "ICON_FAILURE",
        FileNotFoundError(),
    )

    icon = toga.Icon("resources/missing")

    assert icon._impl is not None
    assert icon._impl.interface == toga.Icon.DEFAULT_ICON

    # A warning was printed; allow for windows separators
    assert (
        "WARNING: Can't find icon resources/missing"
        in capsys.readouterr().out.replace("\\", "/")
    )


def test_create_fallback_unloadable(monkeypatch, app, capsys):
    """If a resource exists, but can't be loaded, an error is raised."""
    # Prime the dummy so the app icon cannot be loaded
    monkeypatch.setattr(
        DummyIcon,
        "ICON_FAILURE",
        ValueError("Icon could not be loaded"),
    )

    with pytest.raises(ValueError):
        toga.Icon("resources/sample")


def test_create_fallback_variants(monkeypatch, app, capsys):
    """If a resource with size variants doesn't exist, a fallback icon is used."""
    monkeypatch.setattr(DummyIcon, "SIZES", [32, 72])

    icon = toga.Icon("resources/missing")

    assert icon._impl is not None
    assert icon._impl.interface == toga.Icon.DEFAULT_ICON

    # A warning was printed; allow for windows separators
    assert (
        "WARNING: Can't find icon resources/missing"
        in capsys.readouterr().out.replace("\\", "/")
    )


def test_create_app_icon(monkeypatch, app, capsys):
    """The app icon can be constructed"""
    # Patch the app name to a name that will exist
    monkeypatch.setattr(app, "_app_name", "sample")

    # When running under pytest, code will identify as running as a script

    # Load the app default icon.
    icon = toga.Icon(_APP_ICON)

    # The impl is the app icon.
    assert icon._impl is not None
    assert icon._impl.interface != toga.Icon.DEFAULT_ICON
    assert icon.path == Path("resources/sample")
    assert icon._impl.path == Path(APP_RESOURCES / "sample.png")

    # No warning was printed, as the app icon exists.
    assert capsys.readouterr().out == ""


def test_create_app_icon_missing(monkeypatch, app, capsys):
    """If the app icon is missing, a fallback is used"""
    # When running under pytest, code will identify as running as a script

    # Load the app default icon.
    icon = toga.Icon(_APP_ICON)

    # The impl is the app icon.
    assert icon._impl is not None
    assert icon._impl.interface == toga.Icon.DEFAULT_ICON

    # No warning was printed, as we're running as a script.
    assert capsys.readouterr().out == ""


def test_create_app_icon_non_script(monkeypatch, app, capsys):
    """The icon from the binary is used when running as a packaged binary"""
    # Patch sys.executable so the test looks like it's running as a packaged binary
    monkeypatch.setattr(sys, "executable", "/path/to/App")

    # Load the app default icon
    icon = toga.Icon(_APP_ICON)

    assert isinstance(icon, toga.Icon)
    # App icon path reports as `resources/<app_name>`; impl is the app icon
    assert icon.path == Path("resources/icons")
    assert icon._impl.path == "<APP ICON>"

    # No warning was printed, as we're running as a script.
    assert capsys.readouterr().out == ""


def test_create_app_icon_missing_non_script(monkeypatch, app, capsys):
    """If the icon from binary executable cannot be found, the app icon is reset to the default"""
    # Prime the dummy so the app icon cannot be found
    monkeypatch.setattr(
        DummyIcon,
        "ICON_FAILURE",
        FileNotFoundError(),
    )

    # Patch sys.executable so the test looks like it's running as a packaged binary
    monkeypatch.setattr(sys, "executable", "/path/to/App")

    # Load the app default icon
    icon = toga.Icon(_APP_ICON)

    assert isinstance(icon, toga.Icon)
    # App icon path reports as `resources/<app_name>`; impl is the default toga icon
    assert icon.path == Path("resources/icons")
    assert icon._impl.path == Path(TOGA_RESOURCES / "toga.png")

    # A warning was printed; allow for windows separators
    assert "WARNING: Can't find app icon" in capsys.readouterr().out.replace("\\", "/")


@pytest.mark.parametrize(
    "name, path",
    [
        ("DEFAULT_ICON", "toga"),
        ("OPTION_CONTAINER_DEFAULT_TAB_ICON", "optioncontainer-tab"),
    ],
)
def test_cached_icons(app, name, path):
    """Default icons exist, and are cached."""

    icon = getattr(toga.Icon, name)
    assert icon.path == Path(path)

    # Retrieve the icon a second time; The same instance is returned.
    assert id(getattr(toga.Icon, name)) == id(icon)


def test_deprecated_icons(app):
    """Deprecated icons are still available."""
    with pytest.warns(DeprecationWarning):
        icon = toga.Icon.TOGA_ICON
    assert icon.path == Path("toga")

    # Retrieve the icon a second time; The same instance is returned.
    assert id(toga.Icon.TOGA_ICON) == id(icon)



from travertino.size import at_least

from toga.style.pack import COLUMN, ROW, Pack

from ..utils import ExampleNode, ExampleViewport, assert_layout


def test_beeliza():
    root = ExampleNode(
        "app",
        style=Pack(direction=COLUMN),
        children=[
            ExampleNode(
                "detailedlist",
                style=Pack(flex=1),
                size=(at_least(100), at_least(100)),
            ),
            ExampleNode(
                "box",
                style=Pack(direction=ROW),
                children=[
                    ExampleNode(
                        "input",
                        style=Pack(flex=1, padding=5),
                        size=(at_least(100), 15),
                    ),
                    ExampleNode(
                        "button", style=Pack(padding=5), size=(at_least(40), 10)
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (160, 125),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {"origin": (0, 0), "content": (640, 455)},
                {
                    "origin": (0, 455),
                    "content": (640, 25),
                    "children": [
                        {"origin": (5, 460), "content": (580, 15)},
                        {"origin": (595, 460), "content": (40, 10)},
                    ],
                },
            ],
        },
    )



from travertino.size import at_least

from toga.style.pack import COLUMN, ROW, Pack

from ..utils import ExampleNode, ExampleViewport, assert_layout


def test_row_flex_no_hints():
    """Child in a row layout with flexible containers, but no flex hints, doesn't
    collapse row height."""
    root = ExampleNode(
        "app",
        style=Pack(direction=ROW),
        children=[
            ExampleNode(
                "first",
                style=Pack(flex=1),
            ),
            ExampleNode(
                "second",
                style=Pack(direction=ROW, flex=1),
                children=[
                    ExampleNode(
                        "child",
                        style=Pack(width=50, height=50),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (50, 50),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {
                    "origin": (0, 0),
                    "content": (320, 480),
                },
                {
                    "origin": (320, 0),
                    "content": (320, 480),
                    "children": [
                        {"origin": (320, 0), "content": (50, 50)},
                    ],
                },
            ],
        },
    )


def test_row_flex():
    """Children in a row layout with flexible containers, and an explicit intrinsic
    width doesn't collapse row height."""
    root = ExampleNode(
        "app",
        style=Pack(direction=ROW),
        children=[
            ExampleNode(
                "first",
                style=Pack(flex=1),
                size=(100, 100),
            ),
            ExampleNode(
                "second",
                style=Pack(direction=ROW, flex=1),
                size=(200, 100),
                children=[
                    ExampleNode(  # Explicit size
                        "child 1",
                        style=Pack(width=20, height=20),
                    ),
                    ExampleNode(  # No size, Explicit intrinsic
                        "child 2",
                        style=Pack(),
                        size=(20, 20),
                    ),
                    ExampleNode(  # No size, explicit expanding intrinsic
                        "child 3",
                        style=Pack(),
                        size=(at_least(20), at_least(20)),
                    ),
                    ExampleNode(  # Flexible size, no intrinsic
                        "child 4",
                        style=Pack(flex=1),
                    ),
                    ExampleNode(  # Flexible size with explicit intrinsic
                        "child 5",
                        style=Pack(flex=1),
                        size=(20, 20),
                    ),
                    ExampleNode(  # Flexible size with explicit expanding intrinsic
                        "child 6",
                        style=Pack(flex=1),
                        size=(at_least(20), at_least(20)),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (300, 100),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {
                    "origin": (0, 0),
                    "content": (100, 100),
                },
                {
                    "origin": (100, 0),
                    "content": (200, 100),
                    "children": [
                        {"origin": (100, 0), "content": (20, 20)},
                        {"origin": (120, 0), "content": (20, 20)},
                        {"origin": (140, 0), "content": (20, 100)},
                        {"origin": (160, 0), "content": (60, 100)},
                        {"origin": (220, 0), "content": (20, 20)},
                        {"origin": (240, 0), "content": (60, 100)},
                    ],
                },
            ],
        },
    )


def test_row_flex_insufficient_space():
    """Children in a row layout with flexible containers, but insufficient_space, and an
    explicit intrinsic width doesn't collapse row height."""
    root = ExampleNode(
        "app",
        style=Pack(direction=ROW),
        children=[
            ExampleNode(
                "first",
                style=Pack(flex=1),
                size=(100, 100),
            ),
            ExampleNode(
                "second",
                style=Pack(direction=ROW, flex=1),
                size=(100, 100),
                children=[
                    ExampleNode(  # Explicit size
                        "child 1",
                        style=Pack(width=20, height=20),
                    ),
                    ExampleNode(  # No size, Explicit intrinsic
                        "child 2",
                        style=Pack(),
                        size=(20, 20),
                    ),
                    ExampleNode(  # No size, explicit expanding intrinsic
                        "child 3",
                        style=Pack(),
                        size=(at_least(20), at_least(20)),
                    ),
                    ExampleNode(  # Flexible size, no intrinsic
                        "child 4",
                        style=Pack(flex=1),
                    ),
                    ExampleNode(  # Flexible size with explicit intrinsic
                        "child 5",
                        style=Pack(flex=1),
                        size=(20, 20),
                    ),
                    ExampleNode(  # Flexible size with explicit expanding intrinsic
                        "child 6",
                        style=Pack(flex=1),
                        size=(at_least(20), at_least(20)),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (200, 100),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {
                    "origin": (0, 0),
                    "content": (100, 100),
                },
                {
                    "origin": (100, 0),
                    "content": (100, 100),
                    "children": [
                        {"origin": (100, 0), "content": (20, 20)},
                        {"origin": (120, 0), "content": (20, 20)},
                        {"origin": (140, 0), "content": (20, 100)},
                        {"origin": (160, 0), "content": (0, 100)},
                        {"origin": (160, 0), "content": (20, 20)},
                        {"origin": (180, 0), "content": (20, 100)},
                    ],
                },
            ],
        },
    )


def test_row_flex_insufficient_space_no_flex():
    """Children in a row layout with flexible containers, but insufficient_space for any
    of the flexible content, and an explicit intrinsic width, doesn't collapse row
    height."""
    root = ExampleNode(
        "app",
        style=Pack(direction=ROW),
        children=[
            ExampleNode(
                "first",
                style=Pack(flex=1),
                size=(100, 100),
            ),
            ExampleNode(
                "second",
                style=Pack(direction=ROW, flex=1),
                size=(100, 100),
                children=[
                    # All children are expanding flexible, but the intrinsic minimum
                    # size is greater than what is available.
                    ExampleNode(
                        "child 1",
                        style=Pack(flex=1),
                        size=(at_least(60), 20),
                    ),
                    ExampleNode(
                        "child 2",
                        style=Pack(flex=1),
                        size=(at_least(60), 30),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (220, 100),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {
                    "origin": (0, 0),
                    "content": (100, 100),
                },
                {
                    "origin": (100, 0),
                    "content": (120, 100),
                    "children": [
                        {"origin": (100, 0), "content": (60, 20)},
                        {"origin": (160, 0), "content": (60, 30)},
                    ],
                },
            ],
        },
    )


def test_row_flex_grandchild_min_size():
    """The minimum intrinsic sizes of grandchild of flex row containers are honored."""
    root = ExampleNode(
        "app",
        size=(at_least(0), at_least(0)),
        style=Pack(direction=ROW),
        children=[
            ExampleNode(
                "inner",
                size=(at_least(0), at_least(0)),
                style=Pack(direction=COLUMN, flex=1),
                children=[
                    ExampleNode(
                        "textbox",
                        size=(at_least(100), at_least(100)),
                        style=Pack(flex=1),
                    )
                ],
            ),
            ExampleNode(
                "Button",
                size=(20, at_least(100)),
                style=Pack(),
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (120, 100),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {
                    "origin": (0, 0),
                    "content": (620, 480),
                    "children": [
                        {"origin": (0, 0), "content": (620, 480)},
                    ],
                },
                {
                    "origin": (620, 0),
                    "content": (20, 480),
                },
            ],
        },
    )


def test_column_flex_no_hints():
    """Children in a column layout with flexible containers, but no flex hints, doesn't
    collapse column width."""
    root = ExampleNode(
        "app",
        style=Pack(direction=COLUMN),
        children=[
            ExampleNode(
                "first",
                style=Pack(flex=1),
            ),
            ExampleNode(
                "second",
                style=Pack(direction=COLUMN, flex=1),
                children=[
                    ExampleNode(
                        "child",
                        style=Pack(width=50, height=50),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (50, 50),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {
                    "origin": (0, 0),
                    "content": (640, 240),
                },
                {
                    "origin": (0, 240),
                    "content": (640, 240),
                    "children": [
                        {"origin": (0, 240), "content": (50, 50)},
                    ],
                },
            ],
        },
    )


def test_column_flex():
    """Children in a column layout with flexible containers, and an explicit intrinsic
    width, doesn't collapse column width."""
    root = ExampleNode(
        "app",
        style=Pack(direction=COLUMN),
        children=[
            ExampleNode(
                "first",
                style=Pack(flex=1),
                size=(100, 100),
            ),
            ExampleNode(
                "second",
                style=Pack(direction=COLUMN, flex=1),
                size=(100, 200),
                children=[
                    ExampleNode(  # Explicit size
                        "child 1",
                        style=Pack(width=20, height=20),
                    ),
                    ExampleNode(  # No size, Explicit intrinsic
                        "child 2",
                        style=Pack(),
                        size=(20, 20),
                    ),
                    ExampleNode(  # No size, explicit expanding intrinsic
                        "child 3",
                        style=Pack(),
                        size=(at_least(20), at_least(20)),
                    ),
                    ExampleNode(  # Flexible size, no intrinsic
                        "child 4",
                        style=Pack(flex=1),
                    ),
                    ExampleNode(  # Flexible size with explicit intrinsic
                        "child 5",
                        style=Pack(flex=1),
                        size=(20, 20),
                    ),
                    ExampleNode(  # Flexible size with explicit expanding intrinsic
                        "child 6",
                        style=Pack(flex=1),
                        size=(at_least(20), at_least(20)),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (100, 300),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {
                    "origin": (0, 0),
                    "content": (100, 100),
                },
                {
                    "origin": (0, 100),
                    "content": (100, 200),
                    "children": [
                        {"origin": (0, 100), "content": (20, 20)},
                        {"origin": (0, 120), "content": (20, 20)},
                        {"origin": (0, 140), "content": (100, 20)},
                        {"origin": (0, 160), "content": (100, 60)},
                        {"origin": (0, 220), "content": (20, 20)},
                        {"origin": (0, 240), "content": (100, 60)},
                    ],
                },
            ],
        },
    )


def test_column_flex_insufficient_space():
    """Children in a column layout with flexible containers, but insufficient space to
    accommodate them, and an explicit intrinsic width, doesn't collapse column width."""

    root = ExampleNode(
        "app",
        style=Pack(direction=COLUMN),
        children=[
            ExampleNode(
                "first",
                style=Pack(flex=1),
                size=(100, 100),
            ),
            ExampleNode(
                "second",
                style=Pack(direction=COLUMN, flex=1),
                size=(100, 100),
                children=[
                    ExampleNode(  # Explicit size
                        "child 1",
                        style=Pack(width=20, height=20),
                    ),
                    ExampleNode(  # No size, Explicit intrinsic
                        "child 2",
                        style=Pack(),
                        size=(20, 20),
                    ),
                    ExampleNode(  # No size, explicit expanding intrinsic
                        "child 3",
                        style=Pack(),
                        size=(at_least(20), at_least(20)),
                    ),
                    ExampleNode(  # Flexible size, no intrinsic
                        "child 4",
                        style=Pack(flex=1),
                    ),
                    ExampleNode(  # Flexible size with explicit intrinsic
                        "child 5",
                        style=Pack(flex=1),
                        size=(20, 20),
                    ),
                    ExampleNode(  # Flexible size with explicit expanding intrinsic
                        "child 6",
                        style=Pack(flex=1),
                        size=(at_least(20), at_least(20)),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (100, 200),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {
                    "origin": (0, 0),
                    "content": (100, 100),
                },
                {
                    "origin": (0, 100),
                    "content": (100, 100),
                    "children": [
                        {"origin": (0, 100), "content": (20, 20)},
                        {"origin": (0, 120), "content": (20, 20)},
                        {"origin": (0, 140), "content": (100, 20)},
                        {"origin": (0, 160), "content": (100, 0)},
                        {"origin": (0, 160), "content": (20, 20)},
                        {"origin": (0, 180), "content": (100, 20)},
                    ],
                },
            ],
        },
    )


def test_column_flex_insufficient_space_no_flex():
    """Children in a column layout with flexible containers, but insufficient space for
    any of the flexible content, and an explicit intrinsic width, doesn't collapse
    column width."""

    root = ExampleNode(
        "app",
        style=Pack(direction=COLUMN),
        children=[
            ExampleNode(
                "first",
                style=Pack(flex=1),
                size=(100, 100),
            ),
            ExampleNode(
                "second",
                style=Pack(direction=COLUMN, flex=1),
                size=(100, 100),
                children=[
                    # All children are expanding flexible, but the intrinsic minimum
                    # size is greater than what is available.
                    ExampleNode(
                        "child 1",
                        style=Pack(flex=1),
                        size=(20, at_least(60)),
                    ),
                    ExampleNode(
                        "child 2",
                        style=Pack(flex=1),
                        size=(30, at_least(60)),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (100, 220),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {
                    "origin": (0, 0),
                    "content": (100, 100),
                },
                {
                    "origin": (0, 100),
                    "content": (100, 120),
                    "children": [
                        {"origin": (0, 100), "content": (20, 60)},
                        {"origin": (0, 160), "content": (30, 60)},
                    ],
                },
            ],
        },
    )


def test_column_flex_grandchild_min_size():
    """The minimum intrinsic sizes of grandchild of flex column containers are
    honored."""
    root = ExampleNode(
        "app",
        size=(at_least(0), at_least(0)),
        style=Pack(direction=COLUMN),
        children=[
            ExampleNode(
                "inner",
                size=(at_least(0), at_least(0)),
                style=Pack(direction=ROW, flex=1),
                children=[
                    ExampleNode(
                        "textbox",
                        size=(at_least(100), at_least(100)),
                        style=Pack(flex=1),
                    )
                ],
            ),
            ExampleNode(
                "Button",
                size=(at_least(100), 20),
                style=Pack(),
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (100, 120),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {
                    "origin": (0, 0),
                    "content": (640, 460),
                    "children": [
                        {"origin": (0, 0), "content": (640, 460)},
                    ],
                },
                {
                    "origin": (0, 460),
                    "content": (640, 20),
                },
            ],
        },
    )



from __future__ import annotations

from .base import Source


class ValueSource(Source):
    def __init__(self, value: object = None, accessor: str = "value"):
        super().__init__()
        self.accessor = accessor
        setattr(self, accessor, value)

    def __str__(self) -> str:
        return str(getattr(self, self.accessor, None))

    def __setattr__(self, attr: str, value: object) -> None:
        super().__setattr__(attr, value)
        if attr == getattr(self, "accessor", None):
            self.notify("change", item=value)



from toga.types import Position, Size


def test_position_properties():
    """A Position NamedTuple has X and Y values."""
    p = Position(1, 2)
    assert p.x == 1
    assert p.y == 2
    assert str(p) == "(1, 2)"
    p == (1, 2)  # Tuple equivalence for backwards-compatibility


def test_add_positions():
    """The sum of two Positions combines their X and Y values"""
    assert Position(1, 2) + Position(3, 4) == Position(4, 6)


def test_sub_positions():
    """The difference of two Positions subtracts their X and Y values"""
    assert Position(1, 2) - Position(3, 4) == Position(-2, -2)


def test_size_properties():
    """A Size NamedTuple has a width and height."""
    s = Size(1, 2)
    assert s.width == 1
    assert s.height == 2
    assert str(s) == "(1 x 2)"
    s == (1, 2)  # Tuple equivalence for backwards-compatibility



from unittest.mock import Mock

import pytest

import toga
from toga.validators import Number
from toga_dummy.utils import (
    EventLog,
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
    attribute_value,
)


@pytest.fixture
def validator():
    return Mock(return_value=None)


@pytest.fixture
def widget(validator):
    return toga.TextInput(validators=[validator])


def test_widget_created():
    """A text input can be created."""
    widget = toga.TextInput()
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create TextInput")

    assert not widget.readonly
    assert widget.placeholder == ""
    assert widget.value == ""
    assert widget._on_change._raw is None
    assert widget._on_confirm._raw is None
    assert widget._on_gain_focus._raw is None
    assert widget._on_lose_focus._raw is None
    assert widget.validators == []


def test_create_with_values():
    """A multiline text input can be created with initial values."""
    on_change = Mock()
    on_confirm = Mock()
    on_gain_focus = Mock()
    on_lose_focus = Mock()
    validator1 = Mock(return_value=None)
    validator2 = Mock(return_value=None)

    widget = toga.TextInput(
        value="Some text",
        placeholder="A placeholder",
        readonly=True,
        on_change=on_change,
        on_confirm=on_confirm,
        on_gain_focus=on_gain_focus,
        on_lose_focus=on_lose_focus,
        validators=[validator1, validator2],
    )
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create TextInput")

    assert widget.readonly
    assert widget.placeholder == "A placeholder"
    assert widget.value == "Some text"
    assert widget._on_change._raw == on_change
    assert widget._on_confirm._raw == on_confirm
    assert widget._on_gain_focus._raw == on_gain_focus
    assert widget._on_lose_focus._raw == on_lose_focus
    assert widget.validators == [validator1, validator2]

    # Validators have been invoked with the initial text
    validator1.assert_called_once_with("Some text")
    validator2.assert_called_once_with("Some text")

    # Change handler hasn't been invoked
    on_change.assert_not_called()


@pytest.mark.parametrize(
    "value, expected",
    [
        ("New Text", "New Text"),
        ("", ""),
        (None, ""),
        ("New\nText\n", "New Text "),
        (12345, "12345"),
    ],
)
def test_value(widget, value, expected, validator):
    """The value of the input can be set."""
    # Clear the event log and validator mock
    EventLog.reset()
    validator.reset_mock()

    # Define and set a new change callback
    on_change_handler = Mock()
    widget.on_change = on_change_handler

    widget.value = value
    assert widget.value == expected

    # test backend has the right value
    assert attribute_value(widget, "value") == expected

    # A refresh was performed
    assert_action_performed(widget, "refresh")

    # The validator was invoked
    validator.assert_called_once_with(expected)

    # change handler was invoked
    on_change_handler.assert_called_once_with(widget)


def test_validation_order():
    """Widget value validation is performed in the correct order."""
    results = {}

    def on_change(widget):
        results["valid"] = widget.is_valid

    # Define a validator that only accepts numbers
    text_input = toga.TextInput(on_change=on_change, validators=[Number()])

    # Widget is initially valid with a number
    text_input.value = "1234"

    # Change handler was invoked and results are checked
    assert results["valid"]

    # Widget is invalid with text
    text_input.value = "hello"

    # Change handler was invoked and results are checked
    assert not results["valid"]

    # Widget is valid again with a number
    text_input.value = "1234"

    # Confirm final results are True
    assert results["valid"]


@pytest.mark.parametrize(
    "value, expected",
    [
        (None, False),
        ("", False),
        ("true", True),
        ("false", True),  # Evaluated as a string, this value is true.
        (0, False),
        (1234, True),
    ],
)
def test_readonly(widget, value, expected):
    """The readonly status of the widget can be changed."""
    # Widget is initially not readonly by default.
    assert not widget.readonly

    # Set the readonly status
    widget.readonly = value
    assert widget.readonly == expected

    # Set the widget readonly
    widget.readonly = True
    assert widget.readonly

    # Set the readonly status again
    widget.readonly = value
    assert widget.readonly == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        ("New Text", "New Text"),
        ("", ""),
        (None, ""),
        (12345, "12345"),
    ],
)
def test_placeholder(widget, value, expected):
    """The value of the placeholder can be set."""
    # Clear the event log
    EventLog.reset()

    widget.placeholder = value
    assert widget.placeholder == expected

    # test backend has the right value
    assert attribute_value(widget, "placeholder") == expected

    # A refresh was performed
    assert_action_performed(widget, "refresh")


def test_on_change(widget):
    """The on_change handler can be invoked."""
    # No handler initially
    assert widget._on_change._raw is None

    # Define and set a new callback
    handler = Mock()

    widget.on_change = handler

    assert widget.on_change._raw == handler

    # Invoke the callback
    widget._impl.simulate_change()

    # Callback was invoked
    handler.assert_called_once_with(widget)


def test_on_confirm(widget):
    """The on_confirm handler can be invoked."""
    # No handler initially
    assert widget._on_confirm._raw is None

    # Define and set a new callback
    handler = Mock()

    widget.on_confirm = handler

    assert widget.on_confirm._raw == handler

    # Invoke the callback
    widget._impl.simulate_confirm()

    # Callback was invoked
    handler.assert_called_once_with(widget)


def test_on_gain_focus(widget):
    """The on_gain_focus handler can be invoked."""
    # No handler initially
    assert widget._on_gain_focus._raw is None

    # Define and set a new callback
    handler = Mock()

    widget.on_gain_focus = handler

    assert widget.on_gain_focus._raw == handler

    # Invoke the callback
    widget._impl.simulate_gain_focus()

    # Callback was invoked
    handler.assert_called_once_with(widget)


def test_on_lose_focus(widget):
    """The on_lose_focus handler can be invoked."""
    # No handler initially
    assert widget._on_lose_focus._raw is None

    # Define and set a new callback
    handler = Mock()

    widget.on_lose_focus = handler

    assert widget.on_lose_focus._raw == handler

    # Invoke the callback
    widget._impl.simulate_lose_focus()

    # Callback was invoked
    handler.assert_called_once_with(widget)


def test_change_validators(widget, validator):
    """If the validator list is changed, the new validators are invoked."""
    new_validator1 = Mock(return_value=None)
    new_validator2 = Mock(return_value=None)

    widget.value = "Some text"

    # Clear the validator mock
    validator.reset_mock()

    widget.validators = [new_validator1, new_validator2]

    # Old validator hasn't been invoked
    validator.assert_not_called()

    # Validators have been invoked with the initial text
    new_validator1.assert_called_once_with("Some text")
    new_validator2.assert_called_once_with("Some text")


def test_remove_validators(widget, validator):
    """The validator list can be cleared."""
    widget.value = "Some text"

    # Clear the event log and validator mock
    EventLog.reset()
    validator.reset_mock()

    # Clear the validators
    widget.validators = None

    # Old validator hasn't been invoked
    validator.assert_not_called()


def test_is_valid(widget):
    """Widget validity can be evaluated."""
    validator1 = Mock(return_value=None)
    validator2 = Mock(return_value=None)

    widget.validators = [validator1, validator2]

    # Widget is initially valid
    assert widget.is_valid
    assert_action_not_performed(widget, "set_error")
    assert_action_performed_with(widget, "clear_error")

    # Second validator returns an error
    EventLog.reset()
    validator2.return_value = "Invalid 2"
    widget.value = "hello"  # Triggers validation
    assert not widget.is_valid
    assert_action_performed_with(widget, "set_error", error_message="Invalid 2")
    assert_action_not_performed(widget, "clear_error")

    # First validator also returns an error: this should take priority
    EventLog.reset()
    validator1.return_value = "Invalid 1"
    widget.value = "hello"  # Triggers validation
    assert not widget.is_valid
    assert_action_performed_with(widget, "set_error", error_message="Invalid 1")
    assert_action_not_performed(widget, "clear_error")

    # Change validator order
    EventLog.reset()
    widget.validators = [validator2, validator1]  # Triggers validation
    assert not widget.is_valid
    assert_action_performed_with(widget, "set_error", error_message="Invalid 2")
    assert_action_not_performed(widget, "clear_error")

    # Make the validators pass again
    EventLog.reset()
    validator1.return_value = None
    validator2.return_value = None
    widget.value = "hello"  # Triggers validation
    assert widget.is_valid
    assert_action_not_performed(widget, "set_error")
    assert_action_performed_with(widget, "clear_error")



import asyncio
import sys
from pathlib import Path
from unittest.mock import Mock

import pytest

import toga
from toga.platform import get_platform_factory
from toga_dummy.documents import Document as DummyDocument
from toga_dummy.utils import assert_action_performed


class ExampleDocument(toga.Document):
    def __init__(self, path, app):
        super().__init__(path=path, document_type="Example Document", app=app)

    def create(self):
        self.main_window = toga.DocumentMainWindow(self)

    def read(self):
        self.content = self.path


def test_create_no_cmdline(monkeypatch):
    """A document app can be created with no command line."""
    monkeypatch.setattr(sys, "argv", ["app-exe"])

    app = toga.DocumentApp(
        "Test App",
        "org.beeware.document-app",
        document_types={"foobar": ExampleDocument},
    )
    app.main_loop()

    assert app._impl.interface == app
    assert_action_performed(app, "create DocumentApp")

    assert app.document_types == {"foobar": ExampleDocument}
    assert app.documents == []


def test_create_with_cmdline(monkeypatch):
    """If a document is specified at the command line, it is opened."""
    monkeypatch.setattr(sys, "argv", ["app-exe", "/path/to/filename.foobar"])

    app = toga.DocumentApp(
        "Test App",
        "org.beeware.document-app",
        document_types={"foobar": ExampleDocument},
    )
    app.main_loop()

    assert app._impl.interface == app
    assert_action_performed(app, "create DocumentApp")

    assert app.document_types == {"foobar": ExampleDocument}
    assert len(app.documents) == 1
    assert isinstance(app.documents[0], ExampleDocument)

    # Document content has been read
    assert app.documents[0].content == Path("/path/to/filename.foobar")

    # Document window has been created and shown
    assert_action_performed(app.documents[0].main_window, "create Window")
    assert_action_performed(app.documents[0].main_window, "show")


def test_create_with_unknown_document_type(monkeypatch):
    """If the document specified at the command line is an unknown type, an exception is
    raised."""
    monkeypatch.setattr(sys, "argv", ["app-exe", "/path/to/filename.unknown"])

    with pytest.raises(
        ValueError,
        match=r"Don't know how to open documents of type .unknown",
    ):
        toga.DocumentApp(
            "Test App",
            "org.beeware.document-app",
            document_types={"foobar": ExampleDocument},
        )


def test_create_no_document_type():
    """A document app must manage at least one document type."""
    with pytest.raises(
        ValueError,
        match=r"A document must manage at least one document type.",
    ):
        toga.DocumentApp("Test App", "org.beeware.document-app")


def test_close_single_document_app():
    """An app in single document mode closes the app when the window is closed."""
    # Monkeypatch the dummy impl to use single document mode
    DummyDocument.SINGLE_DOCUMENT_APP = True

    # Mock the app, but preserve the factory
    app = Mock()
    app.factory = get_platform_factory()

    doc = ExampleDocument(path=Path("/path/to/doc.txt"), app=app)

    # Window technically was prevented from closing, but the app has been exited.
    # This must be run as a co-routine.
    async def _do_close():
        return await doc.handle_close(Mock())

    assert not asyncio.get_event_loop().run_until_complete(_do_close())
    app.exit.assert_called_once_with()


def test_close_multiple_document_app():
    """An app in multiple document mode doesn't close when the window is closed."""
    # Monkeypatch the dummy impl to use single document mode
    DummyDocument.SINGLE_DOCUMENT_APP = False

    # Mock the app, but preserve the factory
    app = Mock()
    app.factory = get_platform_factory()

    doc = ExampleDocument(path=Path("/path/to/doc.txt"), app=app)

    # Window has closed, but app has not exited.
    # This must be run as a co-routine.
    async def _do_close():
        return await doc.handle_close(Mock())

    assert asyncio.get_event_loop().run_until_complete(_do_close())
    app.exit.assert_not_called()


@pytest.mark.parametrize("is_single_doc_app", [True, False])
def test_no_close(monkeypatch, is_single_doc_app):
    """A document can prevent itself from being closed."""
    # Monkeypatch the dummy impl to set the app mode
    DummyDocument.SINGLE_DOCUMENT_APP = is_single_doc_app

    # Monkeypatch the Example document to prevent closing.
    # Define this as a co-routine to simulate an implementation that called a dialog.
    async def can_close(self):
        return False

    ExampleDocument.can_close = can_close

    # Mock the app, but preserve the factory
    app = Mock()
    app.factory = get_platform_factory()

    doc = ExampleDocument(path=Path("/path/to/doc.txt"), app=app)

    # Window was prevented from closing.
    # This must be run as a co-routine.
    async def _do_close():
        await doc.handle_close(Mock())

    assert not asyncio.get_event_loop().run_until_complete(_do_close())
    app.exit.assert_not_called()



from __future__ import annotations

import asyncio
import importlib.metadata
import signal
import sys
import warnings
import webbrowser
from collections.abc import Iterator
from email.message import Message
from pathlib import Path
from typing import TYPE_CHECKING, Any, MutableSet, Protocol
from warnings import warn
from weakref import WeakValueDictionary

from toga.command import CommandSet
from toga.documents import Document
from toga.handlers import wrapped_handler
from toga.hardware.camera import Camera
from toga.hardware.location import Location
from toga.icons import Icon
from toga.paths import Paths
from toga.platform import get_platform_factory
from toga.screens import Screen
from toga.types import Position, Size
from toga.widgets.base import Widget
from toga.window import OnCloseHandler, Window

if TYPE_CHECKING:
    from toga.icons import IconContentT
    from toga.types import PositionT, SizeT

# Make sure deprecation warnings are shown by default
warnings.filterwarnings("default", category=DeprecationWarning)


class AppStartupMethod(Protocol):
    def __call__(self, app: App, /, **kwargs: Any) -> Widget:
        """The startup method of the app.

        Called during app startup to set the initial main window content.

        :param app: The app instance that is starting.
        :param kwargs: Ensures compatibility with additional arguments introduced in
            future versions.
        :returns: The widget to use as the main window content.
        """


class OnExitHandler(Protocol):
    def __call__(self, app: App, /, **kwargs: Any) -> bool:
        """A handler to invoke when the app is about to exit.

        The return value of this callback controls whether the app is allowed to exit.
        This can be used to prevent the app exiting with unsaved changes, etc.

        :param app: The app instance that is exiting.
        :param kwargs: Ensures compatibility with additional arguments introduced in
            future versions.
        :returns: ``True`` if the app is allowed to exit; ``False`` if the app is not
            allowed to exit.
        """


class BackgroundTask(Protocol):
    def __call__(self, app: App, /, **kwargs: Any) -> object:
        """Code that should be executed as a background task.

        :param app: The app that is handling the background task.
        :param kwargs: Ensures compatibility with additional arguments introduced in
            future versions.
        """


class WindowSet(MutableSet[Window]):
    def __init__(self, app: App):
        """A collection of windows managed by an app.

        A window is automatically added to the app when it is created, and removed when
        it is closed. Adding a window to an App's window set automatically sets the
        :attr:`~toga.Window.app` property of the Window.
        """
        self.app = app
        self.elements: set[Window] = set()

    def add(self, window: Window) -> None:
        if not isinstance(window, Window):
            raise TypeError("Can only add objects of type toga.Window")
        # Silently not add if duplicate
        if window not in self.elements:
            self.elements.add(window)
            window.app = self.app

    def discard(self, window: Window) -> None:
        if not isinstance(window, Window):
            raise TypeError("Can only discard objects of type toga.Window")
        if window not in self.elements:
            raise ValueError(f"{window!r} is not part of this app")
        self.elements.remove(window)

    ######################################################################
    # 2023-10: Backwards compatibility
    ######################################################################

    def __iadd__(self, window: Window) -> WindowSet:
        # The standard set type does not have a += operator.
        warn(
            "Windows are automatically associated with the app; += is not required",
            DeprecationWarning,
            stacklevel=2,
        )
        return self

    def __isub__(self, other: Window) -> WindowSet:
        # The standard set type does have a -= operator, but it takes sets rather than
        # individual items.
        warn(
            "Windows are automatically removed from the app; -= is not required",
            DeprecationWarning,
            stacklevel=2,
        )
        return self

    ######################################################################
    # End backwards compatibility
    ######################################################################

    def __iter__(self) -> Iterator[Window]:
        return iter(self.elements)

    def __contains__(self, value: object) -> bool:
        return value in self.elements

    def __len__(self) -> int:
        return len(self.elements)


class WidgetRegistry:
    # WidgetRegistry is implemented as a wrapper around a WeakValueDictionary, because
    # it provides a mapping from ID to widget. The mapping is weak so the registry
    # doesn't retain a strong reference to the widget, preventing memory cleanup.
    #
    # The lookup methods (__getitem__(), __iter__(), __len()__, keys(), items(), and
    # values()) are all proxied to underlying data store. Private methods exist for
    # internal use, but those methods shouldn't be used by end-users.

    def __init__(self, *args: Any, **kwargs: Any):
        self._registry = WeakValueDictionary(*args, **kwargs)

    def __len__(self) -> int:
        return len(self._registry)

    def __getitem__(self, key: str) -> Widget:
        return self._registry[key]

    def __contains__(self, key: str) -> bool:
        return key in self._registry

    def __iter__(self) -> Iterator[Widget]:
        return self.values()

    def __repr__(self) -> str:
        return f"{{{', '.join(f'{k!r}: {v!r}' for k, v in sorted(self._registry.items()))}}}"

    def items(self) -> Iterator[tuple[str, Widget]]:
        return self._registry.items()

    def keys(self) -> Iterator[str]:
        return self._registry.keys()

    def values(self) -> Iterator[Widget]:
        return self._registry.values()

    # Private methods for internal use
    def _update(self, widgets: list[Widget]) -> None:
        for widget in widgets:
            self._add(widget)

    def _add(self, widget: Widget) -> None:
        if widget.id in self._registry:
            # Prevent adding the same widget twice or adding 2 widgets with the same id
            raise KeyError(f"There is already a widget with the id {widget.id!r}")
        self._registry[widget.id] = widget

    def _remove(self, id: str) -> None:
        del self._registry[id]


class MainWindow(Window):
    _WINDOW_CLASS = "MainWindow"

    def __init__(
        self,
        id: str | None = None,
        title: str | None = None,
        position: PositionT = Position(100, 100),
        size: SizeT = Size(640, 480),
        resizable: bool = True,
        minimizable: bool = True,
        content: Widget | None = None,
        resizeable: None = None,  # DEPRECATED
        closeable: None = None,  # DEPRECATED
    ):
        """Create a new main window.

        :param id: A unique identifier for the window. If not provided, one will be
            automatically generated.
        :param title: Title for the window. Defaults to the formal name of the app.
        :param position: Position of the window, as a :any:`toga.Position` or tuple of
            ``(x, y)`` coordinates, in :ref:`CSS pixels <css-units>`.
        :param size: Size of the window, as a :any:`toga.Size` or tuple of ``(width,
            height)``, in :ref:`CSS pixels <css-units>`.
        :param resizable: Can the window be resized by the user?
        :param minimizable: Can the window be minimized by the user?
        :param content: The initial content for the window.
        :param resizeable: **DEPRECATED** - Use ``resizable``.
        :param closeable: **DEPRECATED** - Use ``closable``.
        """
        super().__init__(
            id=id,
            title=title,
            position=position,
            size=size,
            resizable=resizable,
            closable=True,
            minimizable=minimizable,
            content=content,
            # Deprecated arguments
            resizeable=resizeable,
            closeable=closeable,
        )

    @property
    def _default_title(self) -> str:
        return App.app.formal_name

    @property
    def on_close(self) -> None:
        """The handler to invoke before the window is closed in response to a user
        action.

        Always returns ``None``. Main windows should use :meth:`toga.App.on_exit`,
        rather than ``on_close``.

        :raises ValueError: if an attempt is made to set the ``on_close`` handler.
        """
        return None

    @on_close.setter
    def on_close(self, handler: OnCloseHandler | None) -> None:
        if handler:
            raise ValueError(
                "Cannot set on_close handler for the main window. "
                "Use the app on_exit handler instead."
            )


class DocumentMainWindow(Window):
    def __init__(
        self,
        doc: Document,
        id: str | None = None,
        title: str | None = None,
        position: PositionT = Position(100, 100),
        size: SizeT = Size(640, 480),
        resizable: bool = True,
        minimizable: bool = True,
    ):
        """Create a new document Main Window.

        This installs a default on_close handler that honors platform-specific document
        closing behavior. If you want to control whether a document is allowed to close
        (e.g., due to having unsaved change), override
        :meth:`toga.Document.can_close()`, rather than implementing an on_close handler.

        :param doc: The document being managed by this window
        :param id: The ID of the window.
        :param title: Title for the window. Defaults to the formal name of the app.
        :param position: Position of the window, as a :any:`toga.Position` or tuple of
            ``(x, y)`` coordinates.
        :param size: Size of the window, as a :any:`toga.Size` or tuple of
            ``(width, height)``, in pixels.
        :param resizable: Can the window be manually resized by the user?
        :param minimizable: Can the window be minimized by the user?
        """
        self.doc = doc
        super().__init__(
            id=id,
            title=title,
            position=position,
            size=size,
            resizable=resizable,
            closable=True,
            minimizable=minimizable,
            on_close=doc.handle_close,
        )

    @property
    def _default_title(self) -> str:
        return self.doc.path.name


class App:
    #: The currently running :class:`~toga.App`. Since there can only be one running
    #: Toga app in a process, this is available as a class property via ``toga.App.app``.
    app: App
    _impl: Any
    _camera: Camera
    _location: Location

    def __init__(
        self,
        formal_name: str | None = None,
        app_id: str | None = None,
        app_name: str | None = None,
        *,
        icon: IconContentT | None = None,
        author: str | None = None,
        version: str | None = None,
        home_page: str | None = None,
        description: str | None = None,
        startup: AppStartupMethod | None = None,
        on_exit: OnExitHandler | None = None,
        id: None = None,  # DEPRECATED
        windows: None = None,  # DEPRECATED
    ):
        """Create a new App instance.

        Once the app has been created, you should invoke the
        :meth:`~toga.App.main_loop()` method, which will start the event loop of your
        App.

        :param formal_name: The human-readable name of the app. If not provided, the
            metadata key ``Formal-Name`` must be present.
        :param app_id: The unique application identifier. This will usually be a
            reversed domain name, e.g. ``org.beeware.myapp``. If not provided, the
            metadata key ``App-ID`` must be present.
        :param app_name: The name of the distribution used to load metadata with
            :any:`importlib.metadata`. If not provided, the following will be tried in
            order:

            #. If the ``__main__`` module is contained in a package, that package's name
               will be used.
            #. If the ``app_id`` argument was provided, its last segment will be used.
               For example, an ``app_id`` of ``com.example.my-app`` would yield a
               distribution name of ``my-app``.
            #. As a last resort, the name ``toga``.
        :param icon: The :any:`icon <IconContentT>` for the app. Defaults to
            :attr:`toga.Icon.APP_ICON`.
        :param author: The person or organization to be credited as the author of the
            app. If not provided, the metadata key ``Author`` will be used.
        :param version: The version number of the app.  If not provided, the metadata
            key ``Version`` will be used.
        :param home_page: The URL of a web page for the app. Used in auto-generated help
            menu items. If not provided, the metadata key ``Home-page`` will be used.
        :param description: A brief (one line) description of the app. If not provided,
            the metadata key ``Summary`` will be used.
        :param startup: A callable to run before starting the app.
        :param on_exit: The initial :any:`on_exit` handler.
        :param id: **DEPRECATED** - This argument will be ignored. If you need a
            machine-friendly identifier, use ``app_id``.
        :param windows: **DEPRECATED** – Windows are now automatically added to the
            current app. Passing this argument will cause an exception.
        """
        ######################################################################
        # 2023-10: Backwards compatibility
        ######################################################################
        if id is not None:
            warn(
                "App.id is deprecated and will be ignored. Use app_id instead",
                DeprecationWarning,
                stacklevel=2,
            )

        if windows is not None:
            raise ValueError(
                "The `windows` constructor argument of toga.App has been removed. "
                "Windows are now automatically added to the current app."
            )
        ######################################################################
        # End backwards compatibility
        ######################################################################

        # Initialize empty widgets registry
        self._widgets = WidgetRegistry()

        # Keep an accessible copy of the app singleton instance
        App.app = self

        # We need a distribution name to load app metadata.
        if app_name is None:
            # If the code is contained in appname.py, and you start the app using
            # `python -m appname`, then __main__.__package__ will be an empty string.
            #
            # If the code is contained in appname.py, and you start the app using
            # `python appname.py`, then __main__.__package__ will be None.
            #
            # If the code is contained in appname/__main__.py, and you start the app
            # using `python -m appname`, then __main__.__package__ will be "appname".
            try:
                main_module_pkg = sys.modules["__main__"].__package__
                if main_module_pkg:
                    app_name = main_module_pkg
            except KeyError:
                # If there's no __main__ module, we're probably in a test.
                pass

        # Try deconstructing the distribution name from the app ID
        if (app_name is None) and app_id:
            app_name = app_id.split(".")[-1]

        # If we still don't have a distribution name, fall back to ``toga`` as a
        # last resort.
        if app_name is None:
            app_name = "toga"

        # Try to load the app metadata with our best guess of the distribution name.
        self._app_name = app_name
        try:
            self.metadata = importlib.metadata.metadata(app_name)
        except importlib.metadata.PackageNotFoundError:
            self.metadata = Message()

        # If a formal name has been provided, use it; otherwise, look to
        # the metadata. However, a formal name *must* be provided.
        if formal_name:
            self._formal_name = formal_name
        else:
            self._formal_name = self.metadata.get("Formal-Name")
        if self._formal_name is None:
            raise RuntimeError("Toga application must have a formal name")

        # If an app_id has been provided, use it; otherwise, look to
        # the metadata. However, an app_id *must* be provided
        if app_id:
            self._app_id = app_id
        else:
            self._app_id = self.metadata.get("App-ID")
        if self._app_id is None:
            raise RuntimeError("Toga application must have an app ID")

        # Other metadata may be passed to the constructor, or loaded with importlib.
        if author:
            self._author = author
        else:
            self._author = self.metadata.get("Author", None)

        if version:
            self._version = version
        else:
            self._version = self.metadata.get("Version", None)

        if home_page:
            self._home_page = home_page
        else:
            self._home_page = self.metadata.get("Home-page", None)

        if description:
            self._description = description
        else:
            self._description = self.metadata.get("Summary", None)

        # Get a platform factory.
        self.factory = get_platform_factory()

        # Instantiate the paths instance for this app.
        self._paths = Paths()

        if icon is None:
            self.icon = Icon.APP_ICON
        else:
            self.icon = icon

        self.on_exit = on_exit

        # We need the command set to exist so that startup et al. can add commands;
        # but we don't have an impl yet, so we can't set the on_change handler
        self._commands = CommandSet()

        self._startup_method = startup

        self._main_window: MainWindow | None = None
        self._windows = WindowSet(self)

        self._full_screen_windows: tuple[Window, ...] | None = None

        self._create_impl()

        # Now that we have an impl, set the on_change handler for commands
        self.commands.on_change = self._impl.create_menus

    def _create_impl(self) -> None:
        self.factory.App(interface=self)

    ######################################################################
    # App properties
    ######################################################################

    @property
    def app_name(self) -> str:
        """The name of the distribution used to load metadata with
        :any:`importlib.metadata` (read-only)."""
        return self._app_name

    @property
    def app_id(self) -> str:
        """The unique application identifier (read-only). This will usually be a
        reversed domain name, e.g. ``org.beeware.myapp``.
        """
        return self._app_id

    @property
    def author(self) -> str | None:
        """The person or organization to be credited as the author of the app
        (read-only)."""
        return self._author

    @property
    def description(self) -> str | None:
        """A brief (one line) description of the app (read-only)."""
        return self._description

    @property
    def formal_name(self) -> str:
        """The human-readable name of the app (read-only)."""
        return self._formal_name

    @property
    def home_page(self) -> str | None:
        """The URL of a web page for the app (read-only). Used in auto-generated help
        menu items."""
        return self._home_page

    @property
    def icon(self) -> Icon:
        """The Icon for the app.

        Can be specified as any valid :any:`icon content <IconContentT>`.
        """
        return self._icon

    @icon.setter
    def icon(self, icon_or_name: IconContentT) -> None:
        if isinstance(icon_or_name, Icon):
            self._icon = icon_or_name
        else:
            self._icon = Icon(icon_or_name)

        try:
            self._impl.set_icon(self._icon)
        except AttributeError:
            # The first time the icon is set, it is *before* the impl has been created,
            # so that the app instance can be instantiated with the correct icon.
            pass

    @property
    def id(self) -> str:
        """**DEPRECATED** – Use :any:`app_id`."""
        warn(
            "App.id is deprecated. Use app_id instead", DeprecationWarning, stacklevel=2
        )
        return self._app_id

    @property
    def version(self) -> str | None:
        """The version number of the app (read-only)."""
        return self._version

    @property
    def is_bundled(self) -> bool:
        """Has the app been bundled as a standalone binary, or is it running as a Python script?"""
        return Path(sys.executable).stem not in {
            "python",
            f"python{sys.version_info.major}",
            f"python{sys.version_info.major}.{sys.version_info.minor}",
        }

    ######################################################################
    # App lifecycle
    ######################################################################

    def add_background_task(self, handler: BackgroundTask) -> None:
        """Schedule a task to run in the background.

        Schedules a coroutine or a generator to run in the background. Control
        will be returned to the event loop during await or yield statements,
        respectively. Use this to run background tasks without blocking the
        GUI. If a regular callable is passed, it will be called as is and will
        block the GUI until the call returns.

        :param handler: A coroutine, generator or callable.
        """
        self.loop.call_soon_threadsafe(wrapped_handler(self, handler))

    def exit(self) -> None:
        """Exit the application gracefully.

        This *does not* invoke the ``on_exit`` handler; the app will be immediately
        and unconditionally closed.
        """
        self._impl.exit()

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """The `event loop
        <https://docs.python.org/3/library/asyncio-eventloop.html>`__ of the app's main
        thread (read-only)."""
        return self._impl.loop

    def main_loop(self) -> None:
        """Start the application.

        On desktop platforms, this method will block until the application has exited.
        On mobile and web platforms, it returns immediately.
        """
        # Modify signal handlers to make sure Ctrl-C is caught and handled.
        signal.signal(signal.SIGINT, signal.SIG_DFL)

        self._impl.main_loop()

    @property
    def main_window(self) -> MainWindow | None:
        """The main window for the app."""
        return self._main_window

    @main_window.setter
    def main_window(self, window: MainWindow | None) -> None:
        self._main_window = window
        self._impl.set_main_window(window)

    def _verify_startup(self) -> None:
        if not isinstance(self.main_window, MainWindow):
            raise ValueError(
                "Application does not have a main window. "
                "Does your startup() method assign a value to self.main_window?"
            )

    def _startup(self) -> None:
        # This is a wrapper around the user's startup method that performs any
        # post-setup validation.
        self.startup()
        self._verify_startup()

    def startup(self) -> None:
        """Create and show the main window for the application.

        Subclasses can override this method to define customized startup behavior;
        however, any override *must* ensure the :any:`main_window` has been assigned
        before it returns.
        """
        self.main_window = MainWindow(title=self.formal_name, id="main")

        if self._startup_method:
            self.main_window.content = self._startup_method(self)

        self.main_window.show()

    ######################################################################
    # App resources
    ######################################################################

    @property
    def camera(self) -> Camera:
        """A representation of the device's camera (or cameras)."""
        try:
            return self._camera
        except AttributeError:
            # Instantiate the camera instance for this app on first access
            # This will raise an exception if the platform doesn't implement
            # the Camera API.
            self._camera = Camera(self)
        return self._camera

    @property
    def commands(self) -> CommandSet:
        """The commands available in the app."""
        return self._commands

    @property
    def location(self) -> Location:
        """A representation of the device's location service."""
        try:
            return self._location
        except AttributeError:
            # Instantiate the location service for this app on first access
            # This will raise an exception if the platform doesn't implement
            # the Location API.
            self._location = Location(self)
        return self._location

    @property
    def paths(self) -> Paths:
        """Paths for platform-appropriate locations on the user's file system.

        Some platforms do not allow access to any file system location other than these
        paths. Even when arbitrary file access is allowed, there are preferred locations
        for each type of content.
        """
        return self._paths

    @property
    def screens(self) -> list[Screen]:
        """Returns a list of available screens."""
        return [screen.interface for screen in self._impl.get_screens()]

    @property
    def widgets(self) -> WidgetRegistry:
        """The widgets managed by the app, over all windows.

        Can be used to look up widgets by ID over the entire app (e.g.,
        ``app.widgets["my_id"]``).

        Only returns widgets that are currently part of a layout. A widget that has been
        created, but not assigned as part of window content will not be returned by
        widget lookup.
        """
        return self._widgets

    @property
    def windows(self) -> WindowSet:
        """The windows managed by the app. Windows are automatically added to the app
        when they are created, and removed when they are closed."""
        return self._windows

    ######################################################################
    # App capabilities
    ######################################################################

    def about(self) -> None:
        """Display the About dialog for the app.

        Default implementation shows a platform-appropriate about dialog using app
        metadata. Override if you want to display a custom About dialog.
        """
        self._impl.show_about_dialog()

    def beep(self) -> None:
        """Play the default system notification sound."""
        self._impl.beep()

    def visit_homepage(self) -> None:
        """Open the application's :any:`home_page` in the default browser.

        If the :any:`home_page` is ``None``, this is a no-op.
        """
        if self.home_page is not None:
            webbrowser.open(self.home_page)

    ######################################################################
    # Cursor control
    ######################################################################

    def hide_cursor(self) -> None:
        """Hide cursor from view."""
        self._impl.hide_cursor()

    def show_cursor(self) -> None:
        """Make the cursor visible."""
        self._impl.show_cursor()

    ######################################################################
    # Window control
    ######################################################################

    @property
    def current_window(self) -> Window | None:
        """Return the currently active window."""
        window = self._impl.get_current_window()
        if window is None:
            return None
        return window.interface

    @current_window.setter
    def current_window(self, window: Window) -> None:
        """Set a window into current active focus."""
        self._impl.set_current_window(window)

    ######################################################################
    # Full screen control
    ######################################################################

    def exit_full_screen(self) -> None:
        """Exit full screen mode."""
        if self.is_full_screen:
            self._impl.exit_full_screen(self._full_screen_windows)
            self._full_screen_windows = None

    @property
    def is_full_screen(self) -> bool:
        """Is the app currently in full screen mode?"""
        return self._full_screen_windows is not None

    def set_full_screen(self, *windows: Window) -> None:
        """Make one or more windows full screen.

        Full screen is not the same as "maximized"; full screen mode is when all window
        borders and other window decorations are no longer visible.

        :param windows: The list of windows to go full screen, in order of allocation to
            screens. If the number of windows exceeds the number of available displays,
            those windows will not be visible. If no windows are specified, the app will
            exit full screen mode.
        """
        self.exit_full_screen()
        if windows:
            self._impl.enter_full_screen(windows)
            self._full_screen_windows = windows

    ######################################################################
    # App events
    ######################################################################

    @property
    def on_exit(self) -> OnExitHandler:
        """The handler to invoke if the user attempts to exit the app."""
        return self._on_exit

    @on_exit.setter
    def on_exit(self, handler: OnExitHandler | None) -> None:
        def cleanup(app: App, should_exit: bool) -> None:
            if should_exit or handler is None:
                app.exit()

        self._on_exit = wrapped_handler(self, handler, cleanup=cleanup)

    ######################################################################
    # 2023-10: Backwards compatibility
    ######################################################################

    @property
    def name(self) -> str:
        """**DEPRECATED** – Use :any:`formal_name`."""
        warn(
            "App.name is deprecated. Use formal_name instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._formal_name

    # Support WindowSet __iadd__ and __isub__
    @windows.setter
    def windows(self, windows: WindowSet) -> None:
        if windows is not self._windows:
            raise AttributeError("can't set attribute 'windows'")

    ######################################################################
    # End backwards compatibility
    ######################################################################


class DocumentApp(App):
    def __init__(
        self,
        formal_name: str | None = None,
        app_id: str | None = None,
        app_name: str | None = None,
        *,
        icon: IconContentT | None = None,
        author: str | None = None,
        version: str | None = None,
        home_page: str | None = None,
        description: str | None = None,
        startup: AppStartupMethod | None = None,
        document_types: dict[str, type[Document]] | None = None,
        on_exit: OnExitHandler | None = None,
        id: None = None,  # DEPRECATED
    ):
        """Create a document-based application.

        A document-based application is the same as a normal application, with the
        exception that there is no main window. Instead, each document managed by the
        app will create and manage its own window (or windows).

        :param document_types: Initial :any:`document_types` mapping.
        """
        if document_types is None:
            raise ValueError("A document must manage at least one document type.")

        self._document_types = document_types
        self._documents: list[Document] = []

        super().__init__(
            formal_name=formal_name,
            app_id=app_id,
            app_name=app_name,
            icon=icon,
            author=author,
            version=version,
            home_page=home_page,
            description=description,
            startup=startup,
            on_exit=on_exit,
            id=id,
        )

    def _create_impl(self) -> None:
        self.factory.DocumentApp(interface=self)

    def _verify_startup(self) -> None:
        # No post-startup validation required for DocumentApps
        pass

    @property
    def document_types(self) -> dict[str, type[Document]]:
        """The document types this app can manage.

        A dictionary of file extensions, without leading dots, mapping to the
        :class:`toga.Document` subclass that will be created when a document with that
        extension is opened. The subclass must take exactly 2 arguments in its
        constructor: ``path`` and ``app``.
        """
        return self._document_types

    @property
    def documents(self) -> list[Document]:
        """The list of documents associated with this app."""
        return self._documents

    def startup(self) -> None:
        """No-op; a DocumentApp has no windows until a document is opened.

        Subclasses can override this method to define customized startup behavior.
        """

    def _open(self, path: Path) -> None:
        """Internal utility method; open a new document in this app, and shows the document.

        :param path: The path to the document to be opened.
        :raises ValueError: If the document is of a type that can't be opened. Backends can
            suppress this exception if necessary to preserve platform-native behavior.
        """
        try:
            DocType = self.document_types[path.suffix[1:]]
        except KeyError:
            raise ValueError(f"Don't know how to open documents of type {path.suffix}")
        else:
            document = DocType(path, app=self)
            self._documents.append(document)
            document.show()



from __future__ import annotations

import warnings
from collections.abc import Iterable
from typing import Any, Literal, Protocol, TypeVar

import toga
from toga.handlers import wrapped_handler
from toga.sources import ListSource, Row, Source

from .base import StyleT, Widget

SourceT = TypeVar("SourceT", bound=Source)


class OnPrimaryActionHandler(Protocol):
    def __call__(self, widget: DetailedList, row: Any, /, **kwargs: Any) -> object:
        """A handler to invoke for the primary action.

        :param widget: The DetailedList that was invoked.
        :param row: The current row for the detailed list.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


class OnSecondaryActionHandler(Protocol):
    def __call__(self, widget: DetailedList, row: Any, /, **kwargs: Any) -> object:
        """A handler to invoke for the secondary action.

        :param widget: The DetailedList that was invoked.
        :param row: The current row for the detailed list.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


class OnRefreshHandler(Protocol):
    def __call__(self, widget: DetailedList, /, **kwargs: Any) -> object:
        """A handler to invoke when the detailed list is refreshed.

        :param widget: The DetailedList that was refreshed.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


class OnSelectHandler(Protocol):
    def __call__(self, widget: DetailedList, /, **kwargs: Any) -> object:
        """A handler to invoke when the detailed list is selected.

        :param widget: The DetailedList that was selected.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """


class DetailedList(Widget):
    def __init__(
        self,
        id: str | None = None,
        style: StyleT | None = None,
        data: SourceT | Iterable | None = None,
        accessors: tuple[str, str, str] = ("title", "subtitle", "icon"),
        missing_value: str = "",
        primary_action: str | None = "Delete",
        on_primary_action: OnPrimaryActionHandler | None = None,
        secondary_action: str | None = "Action",
        on_secondary_action: OnSecondaryActionHandler | None = None,
        on_refresh: OnRefreshHandler | None = None,
        on_select: toga.widgets.detailedlist.OnSelectHandler | None = None,
        on_delete: None = None,  # DEPRECATED
    ):
        """Create a new DetailedList widget.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style will be
            applied to the widget.
        :param data: Initial :any:`data` to be displayed in the list.
        :param accessors: The accessors to use to retrieve the data for each item, in
            the form (title, subtitle, icon).
        :param missing_value: The text that will be shown when a row doesn't provide a
            value for its title or subtitle.
        :param on_select: Initial :any:`on_select` handler.
        :param primary_action: The name for the primary action.
        :param on_primary_action: Initial :any:`on_primary_action` handler.
        :param secondary_action: The name for the secondary action.
        :param on_secondary_action: Initial :any:`on_secondary_action` handler.
        :param on_refresh: Initial :any:`on_refresh` handler.
        :param on_delete: **DEPRECATED**; use ``on_primary_action``.
        """
        super().__init__(id=id, style=style)

        ######################################################################
        # 2023-06: Backwards compatibility
        ######################################################################
        if on_delete:
            if on_primary_action:
                raise ValueError("Cannot specify both on_delete and on_primary_action")
            else:
                warnings.warn(
                    "DetailedList.on_delete has been renamed DetailedList.on_primary_action.",
                    DeprecationWarning,
                )
                on_primary_action = on_delete
        ######################################################################
        # End backwards compatibility.
        ######################################################################

        # Prime the attributes and handlers that need to exist when the widget is created.
        self._accessors = accessors
        self._missing_value = missing_value
        self._primary_action = primary_action
        self._secondary_action = secondary_action
        self.on_select = None

        self._data: SourceT | ListSource = None

        self._impl = self.factory.DetailedList(interface=self)

        self.data = data
        self.on_primary_action = on_primary_action
        self.on_secondary_action = on_secondary_action
        self.on_refresh = on_refresh
        self.on_select = on_select

    @property
    def enabled(self) -> Literal[True]:
        """Is the widget currently enabled? i.e., can the user interact with the widget?
        DetailedList widgets cannot be disabled; this property will always return True; any
        attempt to modify it will be ignored.
        """
        return True

    @enabled.setter
    def enabled(self, value: object) -> None:
        pass

    def focus(self) -> None:
        """No-op; DetailedList cannot accept input focus."""
        pass

    @property
    def data(self) -> SourceT | ListSource:
        """The data to display in the table.

        When setting this property:

        * A :any:`Source` will be used as-is. It must either be a :any:`ListSource`, or
          a custom class that provides the same methods.

        * A value of None is turned into an empty ListSource.

        * Otherwise, the value must be an iterable, which is copied into a new
          ListSource. Items are converted as shown :ref:`here <listsource-item>`.
        """
        return self._data

    @data.setter
    def data(self, data: SourceT | Iterable | None) -> None:
        if data is None:
            self._data = ListSource(data=[], accessors=self.accessors)
        elif isinstance(data, Source):
            self._data = data
        else:
            self._data = ListSource(data=data, accessors=self.accessors)

        self._data.add_listener(self._impl)
        self._impl.change_source(source=self._data)

    def scroll_to_top(self) -> None:
        """Scroll the view so that the top of the list (first row) is visible."""
        self.scroll_to_row(0)

    def scroll_to_row(self, row: int) -> None:
        """Scroll the view so that the specified row index is visible.

        :param row: The index of the row to make visible. Negative values refer to the
            nth last row (-1 is the last row, -2 second last, and so on).
        """
        if len(self.data) > 1:
            if row >= 0:
                self._impl.scroll_to_row(min(row, len(self.data)))
            else:
                self._impl.scroll_to_row(max(len(self.data) + row, 0))

    def scroll_to_bottom(self) -> None:
        """Scroll the view so that the bottom of the list (last row) is visible."""
        self.scroll_to_row(-1)

    @property
    def accessors(self) -> tuple[str, str, str]:
        """The accessors used to populate the list (read-only)"""
        return self._accessors

    @property
    def missing_value(self) -> str:
        """The text that will be shown when a row doesn't provide a value for its
        title or subtitle.
        """
        return self._missing_value

    @property
    def selection(self) -> Row | None:
        """The current selection of the table.

        Returns the selected Row object, or :any:`None` if no row is currently selected.
        """
        try:
            return self.data[self._impl.get_selection()]
        except TypeError:
            return None

    @property
    def on_primary_action(self) -> OnPrimaryActionHandler:
        """The handler to invoke when the user performs the primary action on a row of
        the DetailedList.

        The primary action is "swipe left" on platforms that use swipe interactions;
        other platforms may manifest this action in other ways (e.g, a context menu).

        If no ``on_primary_action`` handler is provided, the primary action will be
        disabled in the UI.
        """
        return self._on_primary_action

    @on_primary_action.setter
    def on_primary_action(self, handler: OnPrimaryActionHandler) -> None:
        self._on_primary_action = wrapped_handler(self, handler)
        self._impl.set_primary_action_enabled(handler is not None)

    @property
    def on_secondary_action(self) -> OnSecondaryActionHandler:
        """The handler to invoke when the user performs the secondary action on a row of
        the DetailedList.

        The secondary action is "swipe right" on platforms that use swipe interactions;
        other platforms may manifest this action in other ways (e.g, a context menu).

        If no ``on_secondary_action`` handler is provided, the secondary action will be
        disabled in the UI.
        """
        return self._on_secondary_action

    @on_secondary_action.setter
    def on_secondary_action(self, handler: OnSecondaryActionHandler) -> None:
        self._on_secondary_action = wrapped_handler(self, handler)
        self._impl.set_secondary_action_enabled(handler is not None)

    @property
    def on_refresh(self) -> OnRefreshHandler:
        """The callback function to invoke when the user performs a refresh action
        (usually "pull down") on the DetailedList.

        If no ``on_refresh`` handler is provided, the refresh UI action will be
        disabled.
        """
        return self._on_refresh

    @on_refresh.setter
    def on_refresh(self, handler: OnRefreshHandler) -> None:
        self._on_refresh = wrapped_handler(
            self, handler, cleanup=self._impl.after_on_refresh
        )
        self._impl.set_refresh_enabled(handler is not None)

    @property
    def on_select(self) -> OnSelectHandler:
        """The callback function that is invoked when a row of the DetailedList is selected."""
        return self._on_select

    @on_select.setter
    def on_select(self, handler: toga.widgets.detailedlist.OnSelectHandler) -> None:
        self._on_select = wrapped_handler(self, handler)

    ######################################################################
    # 2023-06: Backwards compatibility
    ######################################################################

    @property
    def on_delete(self) -> OnPrimaryActionHandler:
        """**DEPRECATED**; Use :any:`on_primary_action`"""
        warnings.warn(
            "DetailedList.on_delete has been renamed DetailedList.on_primary_action.",
            DeprecationWarning,
        )
        return self.on_primary_action

    @on_delete.setter
    def on_delete(self, handler: OnPrimaryActionHandler) -> None:
        warnings.warn(
            "DetailedList.on_delete has been renamed DetailedList.on_primary_action.",
            DeprecationWarning,
        )
        self.on_primary_action = handler



import pytest

from toga.sources.accessors import build_accessors, to_accessor


@pytest.mark.parametrize(
    "heading, accessor",
    [
        ("hello", "hello"),
        ("hello", "hello"),
        ("Hello", "hello"),
        ("Hello1", "hello1"),
        ("Hello 1", "hello_1"),
        ("1 Hello", "_1_hello"),
        ("Hello world", "hello_world"),
        ("Hello World", "hello_world"),
        ("Hello World 1", "hello_world_1"),
        ("Hello World 1!", "hello_world_1"),
        ("1 Hello World", "_1_hello_world"),
        ("1 Hello World!", "_1_hello_world"),
        ("Hello!$@# World!^&*(", "hello_world"),
        (" ", "_"),
        # Multiple whitespace characters are collapsed
        ("Hello - World", "hello_world"),
        ("  ", "_"),
    ],
)
def test_to_accessor(heading, accessor):
    """Headings can be converted into accessors."""

    assert to_accessor(heading) == accessor


@pytest.mark.parametrize("heading", ["$*(!&*@&^*&^!", ""])
def test_to_accessor_failures(heading):
    with pytest.raises(
        ValueError,
        match=r"Unable to automatically generate accessor from heading '.*'",
    ):
        to_accessor(heading)


@pytest.mark.parametrize(
    "headings, overrides, accessors",
    [
        # No overrides
        (
            ["First Col", "Second Col", "Third Col"],
            None,
            ["first_col", "second_col", "third_col"],
        ),
        # Explicitly provided accessors
        (
            ["First Col", "Second Col", "Third Col"],
            ["first", "second", "third"],
            ["first", "second", "third"],
        ),
        # Override some accessors
        (
            ["First Col", "Second Col", "Third Col"],
            ["first", "second", None],
            ["first", "second", "third_col"],
        ),
        # Override some accessors using dictionary
        (
            ["First Col", "Second Col", "Third Col"],
            {"First Col": "first", "Second Col": "second"},
            ["first", "second", "third_col"],
        ),
    ],
)
def test_build_accessors(headings, overrides, accessors):
    """Accessors can be constructed from headings with overrides."""
    assert build_accessors(headings, overrides) == accessors


@pytest.mark.parametrize(
    "headings, overrides, error",
    [
        (
            ["First Col", "Second Col", "Third Col"],
            ["first", "second"],
            r"Number of accessors must match number of headings",
        ),
        (
            ["!!", "Second Col", "Third Col"],
            None,
            r"Unable to automatically generate accessor from heading '!!'",
        ),
    ],
)
def test_build_accessor_failure(headings, overrides, error):
    """If an accessor list can't be built, an error is raised."""
    with pytest.raises(ValueError, match=error):
        build_accessors(headings, overrides)



import pytest

from toga.colors import REBECCAPURPLE, rgb
from toga.constants import Baseline, FillRule
from toga.fonts import SYSTEM, SYSTEM_DEFAULT_FONT_SIZE, Font
from toga_dummy.utils import assert_action_performed

REBECCA_PURPLE_COLOR = rgb(102, 51, 153)


def test_begin_path(widget):
    """A begin path operation can be added."""
    draw_op = widget.context.begin_path()

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == "BeginPath()"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        ("begin path", {}),
    ]


def test_close_path(widget):
    """A close path operation can be added."""
    draw_op = widget.context.close_path()

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == "ClosePath()"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        ("close path", {}),
    ]


@pytest.mark.parametrize(
    "kwargs, args_repr, draw_kwargs",
    [
        # Defaults
        (
            {},
            "color=rgb(0, 0, 0), fill_rule=FillRule.NONZERO",
            {"color": rgb(0, 0, 0), "fill_rule": FillRule.NONZERO},
        ),
        # Color as string name
        (
            {"color": REBECCAPURPLE},
            f"color={REBECCA_PURPLE_COLOR}, fill_rule=FillRule.NONZERO",
            {"color": REBECCA_PURPLE_COLOR, "fill_rule": FillRule.NONZERO},
        ),
        # Color as RGB object
        (
            {"color": REBECCA_PURPLE_COLOR},
            f"color={REBECCA_PURPLE_COLOR}, fill_rule=FillRule.NONZERO",
            {"color": REBECCA_PURPLE_COLOR, "fill_rule": FillRule.NONZERO},
        ),
        # Color as RGB string
        (
            {"color": str(REBECCA_PURPLE_COLOR)},
            f"color={REBECCA_PURPLE_COLOR}, fill_rule=FillRule.NONZERO",
            {"color": REBECCA_PURPLE_COLOR, "fill_rule": FillRule.NONZERO},
        ),
        # Color reset with None
        (
            {"color": None},
            "color=rgb(0, 0, 0), fill_rule=FillRule.NONZERO",
            {"color": rgb(0, 0, 0), "fill_rule": FillRule.NONZERO},
        ),
        # Explicit Non-Zero winding
        (
            {"fill_rule": FillRule.NONZERO},
            "color=rgb(0, 0, 0), fill_rule=FillRule.NONZERO",
            {"color": rgb(0, 0, 0), "fill_rule": FillRule.NONZERO},
        ),
        # Even-Odd winding
        (
            {"fill_rule": FillRule.EVENODD},
            "color=rgb(0, 0, 0), fill_rule=FillRule.EVENODD",
            {"color": rgb(0, 0, 0), "fill_rule": FillRule.EVENODD},
        ),
        # All args
        (
            {"color": REBECCAPURPLE, "fill_rule": FillRule.EVENODD},
            f"color={REBECCA_PURPLE_COLOR}, fill_rule=FillRule.EVENODD",
            {"color": REBECCA_PURPLE_COLOR, "fill_rule": FillRule.EVENODD},
        ),
    ],
)
def test_fill(widget, kwargs, args_repr, draw_kwargs):
    """A primitive fill operation can be added."""
    draw_op = widget.context.fill(**kwargs)

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == f"Fill({args_repr})"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        ("fill", draw_kwargs),
    ]

    # All the attributes can be retrieved.
    for attr, value in draw_kwargs.items():
        assert getattr(draw_op, attr) == value


@pytest.mark.parametrize(
    "kwargs, args_repr, draw_kwargs",
    [
        # Defaults
        (
            {},
            "color=rgb(0, 0, 0), line_width=2.0, line_dash=None",
            {"color": rgb(0, 0, 0), "line_width": 2.0, "line_dash": None},
        ),
        # Color as string name
        (
            {"color": REBECCAPURPLE},
            f"color={REBECCA_PURPLE_COLOR}, line_width=2.0, line_dash=None",
            {"color": REBECCA_PURPLE_COLOR, "line_width": 2.0, "line_dash": None},
        ),
        # Color as RGB object
        (
            {"color": REBECCA_PURPLE_COLOR},
            f"color={REBECCA_PURPLE_COLOR}, line_width=2.0, line_dash=None",
            {"color": REBECCA_PURPLE_COLOR, "line_width": 2.0, "line_dash": None},
        ),
        # Color as RGB string
        (
            {"color": str(REBECCA_PURPLE_COLOR)},
            f"color={REBECCA_PURPLE_COLOR}, line_width=2.0, line_dash=None",
            {"color": REBECCA_PURPLE_COLOR, "line_width": 2.0, "line_dash": None},
        ),
        # Color reset with None
        (
            {"color": None},
            "color=rgb(0, 0, 0), line_width=2.0, line_dash=None",
            {"color": rgb(0, 0, 0), "line_width": 2.0, "line_dash": None},
        ),
        # Line width
        (
            {"line_width": 4.5},
            "color=rgb(0, 0, 0), line_width=4.5, line_dash=None",
            {"color": rgb(0, 0, 0), "line_width": 4.5, "line_dash": None},
        ),
        # Line dash
        (
            {"line_dash": [2, 7]},
            "color=rgb(0, 0, 0), line_width=2.0, line_dash=[2, 7]",
            {"color": rgb(0, 0, 0), "line_width": 2.0, "line_dash": [2, 7]},
        ),
        # All args
        (
            {"color": REBECCAPURPLE, "line_width": 4.5, "line_dash": [2, 7]},
            f"color={REBECCA_PURPLE_COLOR}, line_width=4.5, line_dash=[2, 7]",
            {"color": REBECCA_PURPLE_COLOR, "line_width": 4.5, "line_dash": [2, 7]},
        ),
    ],
)
def test_stroke(widget, kwargs, args_repr, draw_kwargs):
    """A primitive stroke operation can be added."""
    draw_op = widget.context.stroke(**kwargs)

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == f"Stroke({args_repr})"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        ("stroke", draw_kwargs),
    ]

    # All the attributes can be retrieved.
    for attr, value in draw_kwargs.items():
        assert getattr(draw_op, attr) == value


def test_move_to(widget):
    """A move to operation can be added."""
    draw_op = widget.context.move_to(10, 20)

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == "MoveTo(x=10, y=20)"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        ("move to", {"x": 10, "y": 20}),
    ]

    # All the attributes can be retrieved.
    assert draw_op.x == 10
    assert draw_op.y == 20


def test_line_to(widget):
    """A line to operation can be added."""
    draw_op = widget.context.line_to(10, 20)

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == "LineTo(x=10, y=20)"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        ("line to", {"x": 10, "y": 20}),
    ]

    # All the attributes can be retrieved.
    assert draw_op.x == 10
    assert draw_op.y == 20


def test_bezier_curve_to(widget):
    """A Bézier curve to operation can be added."""
    draw_op = widget.context.bezier_curve_to(10, 20, 30, 40, 50, 60)

    assert_action_performed(widget, "redraw")
    assert (
        repr(draw_op) == "BezierCurveTo(cp1x=10, cp1y=20, cp2x=30, cp2y=40, x=50, y=60)"
    )

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        (
            "bezier curve to",
            {"cp1x": 10, "cp1y": 20, "cp2x": 30, "cp2y": 40, "x": 50, "y": 60},
        ),
    ]

    # All the attributes can be retrieved.
    assert draw_op.cp1x == 10
    assert draw_op.cp1y == 20
    assert draw_op.cp2x == 30
    assert draw_op.cp2y == 40
    assert draw_op.x == 50
    assert draw_op.y == 60


def test_quadratic_curve_to(widget):
    """A Quadratic curve to operation can be added."""
    draw_op = widget.context.quadratic_curve_to(10, 20, 30, 40)

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == "QuadraticCurveTo(cpx=10, cpy=20, x=30, y=40)"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        (
            "quadratic curve to",
            {"cpx": 10, "cpy": 20, "x": 30, "y": 40},
        ),
    ]

    # All the attributes can be retrieved.
    assert draw_op.cpx == 10
    assert draw_op.cpy == 20
    assert draw_op.x == 30
    assert draw_op.y == 40


@pytest.mark.parametrize(
    "kwargs, args_repr, draw_kwargs",
    [
        # Defaults
        (
            {"x": 10, "y": 20, "radius": 30},
            "x=10, y=20, radius=30, startangle=0.000, endangle=6.283, anticlockwise=False",
            {
                "x": 10,
                "y": 20,
                "radius": 30,
                "startangle": 0.0,
                "endangle": pytest.approx(6.283185),
                "anticlockwise": False,
            },
        ),
        # Start angle
        (
            {"x": 10, "y": 20, "radius": 30, "startangle": 1.234},
            "x=10, y=20, radius=30, startangle=1.234, endangle=6.283, anticlockwise=False",
            {
                "x": 10,
                "y": 20,
                "radius": 30,
                "startangle": pytest.approx(1.234),
                "endangle": pytest.approx(6.283185),
                "anticlockwise": False,
            },
        ),
        # End angle
        (
            {
                "x": 10,
                "y": 20,
                "radius": 30,
                "endangle": 2.345,
            },
            "x=10, y=20, radius=30, startangle=0.000, endangle=2.345, anticlockwise=False",
            {
                "x": 10,
                "y": 20,
                "radius": 30,
                "startangle": 0.0,
                "endangle": pytest.approx(2.345),
                "anticlockwise": False,
            },
        ),
        # Anticlockwise explicitly False
        (
            {
                "x": 10,
                "y": 20,
                "radius": 30,
                "anticlockwise": False,
            },
            "x=10, y=20, radius=30, startangle=0.000, endangle=6.283, anticlockwise=False",
            {
                "x": 10,
                "y": 20,
                "radius": 30,
                "startangle": 0.0,
                "endangle": pytest.approx(6.283185),
                "anticlockwise": False,
            },
        ),
        # Anticlockwise explicitly False
        (
            {
                "x": 10,
                "y": 20,
                "radius": 30,
                "anticlockwise": True,
            },
            "x=10, y=20, radius=30, startangle=0.000, endangle=6.283, anticlockwise=True",
            {
                "x": 10,
                "y": 20,
                "radius": 30,
                "startangle": 0.0,
                "endangle": pytest.approx(6.283185),
                "anticlockwise": True,
            },
        ),
        # All args
        (
            {
                "x": 10,
                "y": 20,
                "radius": 30,
                "startangle": 1.234,
                "endangle": 2.345,
                "anticlockwise": True,
            },
            "x=10, y=20, radius=30, startangle=1.234, endangle=2.345, anticlockwise=True",
            {
                "x": 10,
                "y": 20,
                "radius": 30,
                "startangle": pytest.approx(1.234),
                "endangle": pytest.approx(2.345),
                "anticlockwise": True,
            },
        ),
    ],
)
def test_arc(widget, kwargs, args_repr, draw_kwargs):
    """An arc operation can be added."""
    draw_op = widget.context.arc(**kwargs)

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == f"Arc({args_repr})"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        ("arc", draw_kwargs),
    ]

    # All the attributes can be retrieved.
    for attr, value in draw_kwargs.items():
        assert getattr(draw_op, attr) == value


@pytest.mark.parametrize(
    "kwargs, args_repr, draw_kwargs",
    [
        # Defaults
        (
            {"x": 10, "y": 20, "radiusx": 30, "radiusy": 40},
            (
                "x=10, y=20, radiusx=30, radiusy=40, "
                "rotation=0.000, startangle=0.000, endangle=6.283, anticlockwise=False"
            ),
            {
                "x": 10,
                "y": 20,
                "radiusx": 30,
                "radiusy": 40,
                "rotation": 0.0,
                "startangle": 0.0,
                "endangle": pytest.approx(6.283185),
                "anticlockwise": False,
            },
        ),
        # Rotation
        (
            {"x": 10, "y": 20, "radiusx": 30, "radiusy": 40, "rotation": 1.234},
            (
                "x=10, y=20, radiusx=30, radiusy=40, "
                "rotation=1.234, startangle=0.000, endangle=6.283, anticlockwise=False"
            ),
            {
                "x": 10,
                "y": 20,
                "radiusx": 30,
                "radiusy": 40,
                "rotation": pytest.approx(1.234),
                "startangle": 0.0,
                "endangle": pytest.approx(6.283185),
                "anticlockwise": False,
            },
        ),
        # Start angle
        (
            {"x": 10, "y": 20, "radiusx": 30, "radiusy": 40, "startangle": 2.345},
            (
                "x=10, y=20, radiusx=30, radiusy=40, "
                "rotation=0.000, startangle=2.345, endangle=6.283, anticlockwise=False"
            ),
            {
                "x": 10,
                "y": 20,
                "radiusx": 30,
                "radiusy": 40,
                "rotation": 0.0,
                "startangle": pytest.approx(2.345),
                "endangle": pytest.approx(6.283185),
                "anticlockwise": False,
            },
        ),
        # End angle
        (
            {"x": 10, "y": 20, "radiusx": 30, "radiusy": 40, "endangle": 3.456},
            (
                "x=10, y=20, radiusx=30, radiusy=40, "
                "rotation=0.000, startangle=0.000, endangle=3.456, anticlockwise=False"
            ),
            {
                "x": 10,
                "y": 20,
                "radiusx": 30,
                "radiusy": 40,
                "rotation": 0.0,
                "startangle": 0.0,
                "endangle": pytest.approx(3.456),
                "anticlockwise": False,
            },
        ),
        # Anticlockwise explicitly False
        (
            {"x": 10, "y": 20, "radiusx": 30, "radiusy": 40, "anticlockwise": False},
            (
                "x=10, y=20, radiusx=30, radiusy=40, "
                "rotation=0.000, startangle=0.000, endangle=6.283, anticlockwise=False"
            ),
            {
                "x": 10,
                "y": 20,
                "radiusx": 30,
                "radiusy": 40,
                "rotation": 0.0,
                "startangle": 0.0,
                "endangle": pytest.approx(6.283185),
                "anticlockwise": False,
            },
        ),
        # Anticlockwise
        (
            {"x": 10, "y": 20, "radiusx": 30, "radiusy": 40, "anticlockwise": True},
            (
                "x=10, y=20, radiusx=30, radiusy=40, "
                "rotation=0.000, startangle=0.000, endangle=6.283, anticlockwise=True"
            ),
            {
                "x": 10,
                "y": 20,
                "radiusx": 30,
                "radiusy": 40,
                "rotation": 0.0,
                "startangle": 0.0,
                "endangle": pytest.approx(6.283185),
                "anticlockwise": True,
            },
        ),
        # All args
        (
            {
                "x": 10,
                "y": 20,
                "radiusx": 30,
                "radiusy": 40,
                "rotation": 1.234,
                "startangle": 2.345,
                "endangle": 3.456,
                "anticlockwise": True,
            },
            (
                "x=10, y=20, radiusx=30, radiusy=40, "
                "rotation=1.234, startangle=2.345, endangle=3.456, anticlockwise=True"
            ),
            {
                "x": 10,
                "y": 20,
                "radiusx": 30,
                "radiusy": 40,
                "rotation": pytest.approx(1.234),
                "startangle": pytest.approx(2.345),
                "endangle": pytest.approx(3.456),
                "anticlockwise": True,
            },
        ),
    ],
)
def test_ellipse(widget, kwargs, args_repr, draw_kwargs):
    """An ellipse operation can be added."""
    draw_op = widget.context.ellipse(**kwargs)

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == f"Ellipse({args_repr})"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        ("ellipse", draw_kwargs),
    ]

    # All the attributes can be retrieved.
    for attr, value in draw_kwargs.items():
        assert getattr(draw_op, attr) == value


def test_rect(widget):
    """A rect operation can be added."""
    draw_op = widget.context.rect(10, 20, 30, 40)

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == "Rect(x=10, y=20, width=30, height=40)"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        ("rect", {"x": 10, "y": 20, "width": 30, "height": 40}),
    ]

    # All the attributes can be retrieved.
    assert draw_op.x == 10
    assert draw_op.y == 20
    assert draw_op.width == 30
    assert draw_op.height == 40


@pytest.mark.parametrize(
    "kwargs, args_repr, draw_kwargs",
    [
        # Defaults
        (
            {"text": "Hello world", "x": 10, "y": 20},
            "text='Hello world', x=10, y=20, font=<Font: system default size system>, "
            "baseline=Baseline.ALPHABETIC",
            {
                "text": "Hello world",
                "x": 10,
                "y": 20,
                "font": Font(SYSTEM, SYSTEM_DEFAULT_FONT_SIZE)._impl,
                "baseline": Baseline.ALPHABETIC,
            },
        ),
        # Baseline
        (
            {"text": "Hello world", "x": 10, "y": 20, "baseline": Baseline.TOP},
            "text='Hello world', x=10, y=20, font=<Font: system default size system>, "
            "baseline=Baseline.TOP",
            {
                "text": "Hello world",
                "x": 10,
                "y": 20,
                "font": Font(SYSTEM, SYSTEM_DEFAULT_FONT_SIZE)._impl,
                "baseline": Baseline.TOP,
            },
        ),
        # Font
        (
            {"text": "Hello world", "x": 10, "y": 20, "font": Font("Cutive", 42)},
            "text='Hello world', x=10, y=20, font=<Font: 42pt Cutive>, "
            "baseline=Baseline.ALPHABETIC",
            {
                "text": "Hello world",
                "x": 10,
                "y": 20,
                "font": Font("Cutive", 42)._impl,
                "baseline": Baseline.ALPHABETIC,
            },
        ),
    ],
)
def test_write_text(widget, kwargs, args_repr, draw_kwargs):
    """A write text operation can be added."""
    draw_op = widget.context.write_text(**kwargs)

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == f"WriteText({args_repr})"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        ("write text", draw_kwargs),
    ]

    # All the attributes can be retrieved.
    assert draw_op.text == draw_kwargs["text"]
    assert draw_op.x == draw_kwargs["x"]
    assert draw_op.y == draw_kwargs["y"]
    assert draw_op.font == draw_kwargs["font"].interface
    assert draw_op.baseline == draw_kwargs["baseline"]


def test_rotate(widget):
    """A rotate operation can be added."""
    draw_op = widget.context.rotate(1.234)

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == "Rotate(radians=1.234)"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        ("rotate", {"radians": pytest.approx(1.234)}),
    ]

    # All the attributes can be retrieved.
    assert draw_op.radians == pytest.approx(1.234)


def test_scale(widget):
    """A scale operation can be added."""
    draw_op = widget.context.scale(1.234, 2.345)

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == "Scale(sx=1.234, sy=2.345)"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        ("scale", {"sx": pytest.approx(1.234), "sy": pytest.approx(2.345)}),
    ]

    # All the attributes can be retrieved.
    assert draw_op.sx == pytest.approx(1.234)
    assert draw_op.sy == pytest.approx(2.345)


def test_translate(widget):
    """A translate operation can be added."""
    draw_op = widget.context.translate(10, 20)

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == "Translate(tx=10, ty=20)"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        ("translate", {"tx": 10, "ty": 20}),
    ]

    # All the attributes can be retrieved.
    assert draw_op.tx == 10
    assert draw_op.ty == 20


def test_reset_transform(widget):
    """A reset transform operation can be added."""
    draw_op = widget.context.reset_transform()

    assert_action_performed(widget, "redraw")
    assert repr(draw_op) == "ResetTransform()"

    # The first and last instructions can be ignored as they are the
    assert widget._impl.draw_instructions[1:-1] == [
        ("reset transform", {}),
    ]


def test_deprecated_usage(widget):
    """Test that deprecated usage of operations raise errors."""
    with pytest.raises(
        RuntimeError,
        match=r"Context\.fill\(\) has been renamed Context\.Fill\(\)\.",
    ):
        with widget.context.fill() as fill:
            fill.line_to(10, 20)

    with pytest.raises(
        RuntimeError,
        match=r"Context\.stroke\(\) has been renamed Context\.Stroke\(\)\.",
    ):
        with widget.context.stroke() as stroke:
            stroke.line_to(10, 20)



from __future__ import annotations

from enum import Enum


class Key(Enum):
    """An enumeration providing a symbolic representation for the characters on
    a keyboard."""

    A = "a"
    B = "b"
    C = "c"
    D = "d"
    E = "e"
    F = "f"
    G = "g"
    H = "h"
    I = "i"  # noqa: E741
    J = "j"
    K = "k"
    L = "l"
    M = "m"
    N = "n"
    O = "o"  # noqa: E741
    P = "p"
    Q = "q"
    R = "r"
    S = "s"
    T = "t"
    U = "u"
    V = "v"
    W = "w"
    X = "x"
    Y = "y"
    Z = "z"

    _0 = "0"
    _1 = "1"
    _2 = "2"
    _3 = "3"
    _4 = "4"
    _5 = "5"
    _6 = "6"
    _7 = "7"
    _8 = "8"
    _9 = "9"

    EXCLAMATION = "!"
    AT = "@"
    HASH = "#"
    DOLLAR = "$"
    PERCENT = "%"
    CARET = "^"
    AMPERSAND = "&"
    ASTERISK = "*"
    OPEN_PARENTHESIS = "("
    CLOSE_PARENTHESIS = ")"

    MINUS = "-"
    UNDERSCORE = "_"
    EQUAL = "="
    PLUS = "+"
    OPEN_BRACKET = "["
    CLOSE_BRACKET = "]"
    OPEN_BRACE = "{"
    CLOSE_BRACE = "}"
    BACKSLASH = "\\"
    PIPE = "|"
    SEMICOLON = ";"
    COLON = ":"

    QUOTE = "'"
    DOUBLE_QUOTE = '"'
    COMMA = ","
    LESS_THAN = "<"
    FULL_STOP = "."
    GREATER_THAN = ">"
    SLASH = "/"
    QUESTION = "?"
    BACK_QUOTE = "`"
    TILDE = "~"

    ESCAPE = "<esc>"
    TAB = "<tab>"
    SPACE = " "

    BACKSPACE = "<backspace>"
    ENTER = "<enter>"

    CAPSLOCK = "<caps lock>"
    SHIFT = "<shift>"

    # Modifiers keys are handled slightly different on every platform
    # However, there is a broad platform-specific precedence for them.
    # Preserve the platform precedence behind generic naming.
    MOD_1 = "<mod 1>"  # CMD on macOS, CTRL on Linux/Windows
    MOD_2 = "<mod 2>"  # OPT on macOS, ALT on Linux/Windows
    MOD_3 = "<mod 3>"  # CTRL on macOS, Flag on Windows, Tux on Linux

    F1 = "<f1>"
    F2 = "<f2>"
    F3 = "<f3>"
    F4 = "<f4>"
    F5 = "<f5>"
    F6 = "<f6>"
    F7 = "<f7>"
    F8 = "<f8>"
    F9 = "<f9>"
    F10 = "<f10>"
    F11 = "<f11>"
    F12 = "<f12>"
    F13 = "<f13>"
    F14 = "<f14>"
    F15 = "<f15>"
    F16 = "<f16>"
    F17 = "<f17>"
    F18 = "<f18>"
    F19 = "<f19>"

    EJECT = "<eject>"

    HOME = "<home>"
    END = "<end>"
    INSERT = "<insert>"
    DELETE = "<delete>"
    PAGE_UP = "<pg up>"
    PAGE_DOWN = "<pg dn>"

    UP = "<up>"
    DOWN = "<down>"
    LEFT = "<left>"
    RIGHT = "<right>"

    NUMLOCK = "<num lock>"
    NUMPAD_0 = "numpad:0"
    NUMPAD_1 = "numpad:1"
    NUMPAD_2 = "numpad:2"
    NUMPAD_3 = "numpad:3"
    NUMPAD_4 = "numpad:4"
    NUMPAD_5 = "numpad:5"
    NUMPAD_6 = "numpad:6"
    NUMPAD_7 = "numpad:7"
    NUMPAD_8 = "numpad:8"
    NUMPAD_9 = "numpad:9"
    NUMPAD_CLEAR = "numpad:clear"
    NUMPAD_DECIMAL_POINT = "numpad:."
    NUMPAD_DIVIDE = "numpad:/"
    NUMPAD_ENTER = "numpad:enter"
    NUMPAD_EQUAL = "numpad:="
    NUMPAD_MINUS = "numpad:-"
    NUMPAD_MULTIPLY = "numpad:*"
    NUMPAD_PLUS = "numpad:+"

    SCROLLLOCK = "<scroll lock>"
    BEGIN = "<begin>"
    MENU = "<menu>"
    PAUSE = "<pause>"

    def is_printable(self) -> bool:
        """Does pressing the key result in a printable character?"""
        return not (self.value.startswith("<") and self.value.endswith(">"))

    def __add__(self, other: Key | str) -> str:
        """Allow two Keys to be concatenated, or a string to be concatenated to a Key.

        Produces a single string definition.

        e.g.,
        ``Toga.Key.MOD_1 + 'a'`` -> ``"<mod 1>a"``
        ``Toga.Key.MOD_1 + Toga.Key.SHIFT + 'a'`` -> ``"<mod 1><shift>a"``
        """
        try:
            # Try Key + Key
            return self.value + other.value
        except AttributeError:
            return self.value + other

    def __radd__(self, other: str) -> str:
        """Same as add."""
        return other + self.value
